/*-
 * Copyright (C)2014 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright (C)2014 @BABOLO http://www.babolo.ru/\n\
@(#)All rights reserved.\n\
@(#)V.M "VERS;
static const char rcsid[] = "$Id: recobe.c,v 1.18 2016/04/29 20:17:27 babolo Exp $";
#endif /* not lint */

#define BLIN_COMPAT  3
#define MIFE_COMPAT  4
#define Bpars_COMPAT 3

#define RECOBE_INTERNAL 1

#include <sys/types.h>
#include <sys/event.h>
#include <sys/stat.h>
#include <sysexits.h>
#include <strings.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <babolo/BLINflag.h>
#include <babolo/parser.h>
#include <mife.h>
#include "recobe.h"

int  recobe_TestOrRevert __P((const char *p));
int  recobe_PrintAndMake __P((const char *p));
int  recobe_PrintNoRecod __P((const char *p));
int  recobe_WriteAndMake __P((const char *p));
int  recobe_G711DiffConv __P((const char *p));
int  recobe_PrintFilling __P((const char *p));
int  recobe_InterMediaTo __P((const char *p));
int  recobe_EtalonCompar __P((const char *p));
int  recobe_ReCalculator __P((const char *p));
int  recobe_ReCoordinate __P((const char *p));
int  recobe_PrintWideStr __P((const char *p));
void recobe_p256         __P(( const char *template
                             , const char *from
                             , const char *to
                             , u_int16_t domain
                             , const u_char array[256]
                            ));

static void
/*****************************************************************************************************/
usage(int ex) {                                                                                  /****
 *****************************************************************************************************/
    fprintf( stderr
           , "recobe @BABOLO V.M "VERS"  "DATE" http://www.babolo.ru/\n"
             "Usage: recobe [-options][ infile_or_string[ oufile]]\n"
             "   -c cnv          convert byte-to-byte infile to oufile\n"
             "   -e str          substitute string if conversion error\n"
             "   -E -X XX        use XX as env name for opt's arg\n"
             "   -f              full control log\n"
             "   -k cnv          convert text-to-text infile to oufile\n"
             "   -l              ?\n"
             "   -o ord          byte orer: B, L\n"
             "   -p              print printables as symbols\n"
             "   -P              fancy print\n"
             "   -q              silent\n"
             "   -s code         control code\n"
             "   -S              strict recoding\n"
             "   -t XY[name]     print table from binary stdin\n"
             "   -v[v[v...]]     verbose\n"
             "   -z TEXT         parce TEXT as arguments continuation\n"
             "   -Z TEXT         parce TEXT as arguments continuation with env substitution\n"
             "   -= mode         internal diagnostic\n"
             "      mode: 0             blin_ctl(BLIN_CTL_DUMP)\n"
             "            1             babolo_optest(OptionsString)\n"
             "            2             babolo_optext(OptionsDescriptor)\n"
             "            3             babolo_dumpopts(OptionsDescriptor)\n"
             "            4[D]          inconsistent reverse, D - domain (1 TEXT, 2 G711)\n"
             "            5[Y[Z]]       print table(s); Y, Z code names\n"
             "            S[Y[Z]]       write table(s); Y, Z code names\n"
             "            6[A[B[name]]] print norecoding table nameA2B[]\n"
             "            7{u|A|Au|uA}  inconsistent G711u, G711A, A=>u, u=>a\n"
             "            8[D]          count, D - domain\n"
             "            9[Y[Z]]       test intermediators; Y, Z code names\n"
             "            E[Y[Z]]       compare with etalons\n"
             "   -?              this text\n"
           )
    ;
    exit(ex);
}

int
/*****************************************************************************************************
 *****************************************************************************************************/
main(int argc, char **argv) {                                                                    /****
 *****************************************************************************************************
 *****************************************************************************************************/
    const char       *o, *Tflag = NULL, *tflag = NULL, *cflag = NULL, *kflag = NULL;
    const char       *flar = "c:e:fk:lo:pPqs:St:v=:?";
    const char       *subst = "?";
    FILE             *in = stdin, *ou = stdout;
    int               c, i, k, code, ex = EX_OK;
    u_char           *table = NULL;
    mife_descriptor  *ifile = NULL;            /*                                         */
    u_char            order = 'L';
    u_char            obuf[16];
    off_t             ioff;
    u_char           *inin;
    babolo_opts      *bos;
    ssize_t           j, insz;
    u_char           *z;
    u_int64_t         wchr;

    blin_ctl(BLIN_CTL_FSET, BLIN_BIT0);
    if  (!(bos = babolo_openopts(Bpars_ESYM | Bpars_SYME, 'Z'))) {
        ifBLIN_QW0("openopts");
        ex = EX_SOFTWARE;
        goto out;
    }
    if  (!!babolo_setopts(bos, Bpars_NONU, argc, argv, flar)) {
        ifBLIN_QW0("setopts");
        ex = EX_SOFTWARE;
        goto out;
    }
    code = 0;
    for (tflag = 0;;) {
        c = babolo_getopts(bos);
        if  (!c) break;
        switch (c) {
        case 'c': cflag = babolo_getoptsarg(bos);
                ; break;
        case 'e': subst = babolo_getoptsarg(bos);
                ; break;
        case 'f': blin_internal_flags |= RECOBE_FULLCHECK;
                ; break;
        case 'k': kflag = babolo_getoptsarg(bos);
                ; break;
        case 'l': blin_internal_flags |= RECOBE_PRELINERA;
                ; break;
        case 'o': o = babolo_getoptsarg(bos);
                ; switch(o[0]) {
                  case 'b':
                  case 'l':
                      order = o[0] & 0x00DF;
                      break;
                  case 'B':
                  case 'L':
                      order = o[0];
                      break;
                  default:
                      ifBLIN_QX0("Illegal argument %s of -o option", o);
                  }
                ; break;
        case 'p': blin_internal_flags |= RECOBE_PRINTABLE;
                ; break;
        case 'P': blin_internal_flags |= RECOBE_PRINTATLE;
                ; break;
        case 'q': blin_ctl(BLIN_CTL_FRST, BLIN_MASK);
                ; break;
        case 's': code = babolo_getoptsnum(bos, 0);
                ; break;
        case 'S': blin_internal_flags |= RECOBE_STRICT;
                ; break;
        case 't': tflag = babolo_getoptsarg(bos);
                ; break;
        case 'v': BLIN_VERBOSE(blin_internal_flags);
                ; break;
        case '=': switch((Tflag = babolo_getoptsarg(bos))[0]) {
                  case '0': blin_ctl(BLIN_CTL_DUMP);      exit(EX_OK);
                  case '1': babolo_optest(flar);          exit(EX_OK);
                  case '2': babolo_optext(bos);           exit(EX_OK);
                  case '3': babolo_dumpopts(bos);         exit(EX_OK);
                  }
                ; break;
        case '?': usage(EX_OK);
        default : ifBLIN_QX0("Illegal flag -%c", bos->c);
                ; usage(EX_USAGE);
    }   }
    if  (Tflag) {
        switch (Tflag[0]) {
        case '4': ex = recobe_TestOrRevert(++Tflag); break;
        case '5': ex = recobe_PrintAndMake(++Tflag); break;
        case 'S': ex = recobe_WriteAndMake(++Tflag); break;
        case '6': ex = recobe_PrintNoRecod(++Tflag); break;
        case '7': ex = recobe_G711DiffConv(++Tflag); break;
        case '8': ex = recobe_PrintFilling(++Tflag); break;
        case '9': ex = recobe_InterMediaTo(++Tflag); break;
        case 'E': ex = recobe_EtalonCompar(++Tflag); break;
        case 'R': ex = recobe_ReCalculator(++Tflag); break;
        case 'C': ex = recobe_ReCoordinate(++Tflag); break;
        case 'W': ex = recobe_PrintWideStr(++Tflag); break;
        }
    } else if (!!code) {
        j = 0;
        if  (!(z = babolo_hextrin(babolo_getargs(bos), &j))) {
            ifBLIN_QW0("babolo_hextrin");
            ex = EX_IOERR;
            goto out;
        }
        ifBLIN_QO1 blin_dumb(1, z, j);
        if  (0 > code) j = -j;
        switch ((0 > code) ? -code : code) {
        case 1: printf("csum1 = %02X\n", recobe_csum1(z, j));
                break;
        case 2: printf("crc24 = %06X\n", recobe_crc24(z, j));
                break;
        case 3: printf("crc32 = %08X\n", recobe_crc32(z, j));
                break;
        default:
                ifBLIN_QX0("Illegal code %d", code);
        }
    } else {
        if  (!(ifile = mife_init(BLIN_GEN0 | MIFE_SLID))) {
            ifBLIN_QW0("mife_init");
            ex = EX_IOERR;
            goto out;
        }
        if  (!!(o = babolo_getargs(bos))) {
            if  (0 > mife_ctl(ifile, MIFE_CTLFILE, o)) {
                ifBLIN_QW0("MIFE_CTLFILE=%s~", o);
                ex = EX_IOERR;
                goto out;
            }
        } else {
            if  (!fflush(stdin)) {
                ifBLIN_QW0("fflush");
                ex = EX_IOERR;
                goto out;
            }
            if  (0 > mife_ctl(ifile, MIFE_CTLFDSC, fileno(stdin))) {
                ifBLIN_QW0("MIFE_CTLFDSC");
                ex = EX_IOERR;
                goto out;
        }   }
        if  ((o = babolo_getargs(bos)) && !(ou = fopen(o, "w"))) {
            ifBLIN_QW0("fopen ou %s", o);
            ex = EX_IOERR;
            goto out;
        }
        if  (!!cflag) {
            if  (!(table = recobe_collection.get(cflag, 0))) {
                ifBLIN_QX0("No recoding %s", argv[1]);
                ex = EX_USAGE;
                usage(ex);
            }
            for (ioff = 0;; ++ioff) {
                if  (0 > (insz = mife_ctl(ifile, MIFE_CTLREAD, ioff, (ssize_t)1))) {
                    ifBLIN_QW0("MIFE_CTLREAD");
                    ex = EX_IOERR;
                    goto out;
                }
                if  (!insz) break;
                if  (!(inin = mife_point(ifile, MIFE_POIWIND, ioff, (ssize_t)1))) {
                    ifBLIN_QW0("MIFE_POIWIND");
                    ex = EX_IOERR;
                    goto out;
                }
                i = *inin;
                i &= 0xFF;
                j = table[i];
                if  (!!i & !j) {
                    if  (!fflush(stdin)) {
                        ifBLIN_QW0("fflush");
                        ex = EX_IOERR;
                        goto out;
                    };
                    printf("%s", subst);
                    if  (!fflush(stdin)) {
                        ifBLIN_QW0("fflush");
                        ex = EX_IOERR;
                        goto out;
                }   }
                putc(j, ou);
            }
            free(table);
        } else if (!!kflag) {
            if  ((kflag[0] == 'u') && (kflag[1] >= '0') && (kflag[1] <= '9')) {
                i = c = kflag[1] - '0' + 1;
                if  (order == 'B') i = -c;
                for (ioff = 0;;) {
                    if  (0 > (insz = mife_ctl(ifile, MIFE_CTLREAD, ioff, (ssize_t)8))) {
                        ifBLIN_QW0("MIFE_CTLREAD");
                        ex = EX_IOERR;
                        goto out;
                    }
                    if  (!insz) break;
                    if  (!(inin = mife_point(ifile, MIFE_POIWIND, ioff, (ssize_t)1))) {
                        ifBLIN_QW0("MIFE_POIWIND");
                        ex = EX_IOERR;
                        goto out;
                    }
                    recobe_store(obuf, i, recobe_utoW(inin, insz, &j));
                    if  (j > 0) {
                        ioff += j;
                        if  (1 > fwrite(obuf, c, 1, ou)) {
                            ifBLIN_QW0("fwrite");
                            ex = EX_IOERR;
                            goto out;
                        }
                    } else {
                        ioff -= j;
                        for (j = 0; !!subst[j]; ++j) {
                            recobe_store(obuf, i, subst[j]);
                            if  (1 > fwrite(obuf, c, 1, ou)) {
                                ifBLIN_QW0("fwrite");
                                ex = EX_IOERR;
                                goto out;
                }   }   }   }
            } else if ((kflag[1] == 'k') && (kflag[0] >= '0') && (kflag[0] <= '9')) {
                k = c = kflag[0] - '0' + 1;
                if  (c > 8) k = 8;
                i = k;
                if  (order == 'B') i = -k;
                for (ioff = 0;; ioff += c) {
                    if  (0 > (insz = mife_ctl(ifile, MIFE_CTLREAD, ioff, c))) {
                        ifBLIN_QW0("MIFE_CTLREAD");
                        ex = EX_IOERR;
                        goto out;
                    }
                    if  (!insz) break;                                    /* XXXX if (insz < c) XXXX */
                    if  (!(inin = mife_point(ifile, MIFE_POIWIND, ioff, c))) {
                        ifBLIN_QW0("MIFE_POIWIND");
                        ex = EX_IOERR;
                        goto out;
                    }
                    wchr = recobe_load(inin, i);
                    if  ((wchr > 65535) || ((c > 8) && !!inin[8]) || ((c > 9) && !!inin[9])) {
                        printf("%s", subst);
                    } else {
                        o = recobe_wd[wchr];
                        if  (  (!*o && !!wchr)
                            || (  (blin_internal_flags & RECOBE_STRICT)
                               && (!!o[1] || (recobe_k1[o[0] & 0x00FF] != wchr))
                            )  ) {
                            printf("%s", subst);
                        } else {
                            printf("%s", o);
            }   }   }   }
        } else if (!!tflag) {
            char f[1] = "0", t[1] = "0";

            f[0] = tflag[0];
            t[0] = tflag[1];
            ioff = 0;
            if  (256 > (insz = mife_ctl(ifile, MIFE_CTLREAD, ioff, (ssize_t)256))) {
                ifBLIN_QW0("MIFE_CTLREAD");
                ex = EX_IOERR;
                goto out;
            }
            if  (!(inin = mife_point(ifile, MIFE_POIWIND, ioff, (ssize_t)256))) {
                ifBLIN_QW0("MIFE_POIWIND");
                ex = EX_IOERR;
                goto out;
            }
            recobe_p256(tflag + 2, f, t, RECOBE_TEXT, inin);
        }
        fclose(in);
        fclose(ou);
    }
out:
    exit(ex);
}
